#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "xwchar.h"
#include "text.h"
#include "image.h"
#include "frame.h"
#include "href.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif

static AlignInfo *
aligninfo_new(TAGS tag, int align)
{
	AlignInfo *p;
    if ((p = malloc(sizeof(*p))) == NULL) {
        perror("aligninfo_new");
        exit(1);
    }
	p->next = NULL;
	p->tag  = tag;
	p->align = align;
	return p;
}

static void
aligninfo_delete(AlignInfo *p)
{
	if (p)
		free(p);
}

static void
aligninfolist_delete(AlignInfo *p)
{
	AlignInfo *q;
	while(p) {
		q = p;
		p = p->next;
		free(p);
	}
}

static ColorList *
colorlist_new(void)
{
	ColorList *p;
    if ((p = malloc(sizeof(*p))) == NULL) {
        perror("colorlist_new");
        exit(1);
    }
	return p;
}

static void
colorlistlist_delete(Frame *f, ColorList *p)
{
	ColorList *q;
	while(p) {
		q = p;
		XFreeColors(f->dpy, f->colormap, &p->pixel, 1, 0);
		p = p->next;
		free(p);
	}
}

void 
frame_alloc_color(Frame *f, char *name, u_long *ret)
{
	XColor scr, exact;
	XrmQuark n;
	ColorList *p;

	p = f->pixels;
	while (p != NULL) {
		if ((n = XrmStringToQuark(name)) == p->name) {
			*ret = p->pixel;
			return;
		}
		p = p->next;
	}	
	if (XAllocNamedColor(f->dpy, f->colormap, name, &scr, &exact)) {
		p = colorlist_new();
		p->name = n;
		*ret = p->pixel = scr.pixel;
		p->next = f->pixels;
		f->pixels = p;
	}
}

static FrameInfo *
frameinfo_new(Frame *f, int off_x, int off_y)
{
	FrameInfo *p;
	FrameInfo *fi = f->finfo;

    if ((p = malloc(sizeof(*p))) == NULL) {
        perror("frameinfo_new");
        exit(1);
    }
	
	p->next = NULL;
	p->tag  = TAG_none;

	p->after_lf  = 1;
	p->pre       = TAG_CLOSE;
	p->paragraph = 0;
	p->align   = aligninfo_new(0,(fi == NULL)?ALIGN_NONE:(fi->align->align));
	p->li_cnt  = 0;
	p->li_level = (fi == NULL) ? 0: fi->li_level;
	
	p->width = ((fi == NULL)? f->width: (fi->width - CUR_X)) - off_x;
	p->off_x = off_x + ((fi == NULL)?0: (fi->off_x + CUR_X));
	p->off_y = off_y + ((fi == NULL)?0: (fi->off_y + fi->cur_y));
	
	p->cur_x = 0;
	p->cur_y = 0;
	
	p->left  = htmlobj_new();
	p->right = htmlobj_new();
	p->left_image_margin  = 0;
	p->right_image_height = 0;
	p->right_image_margin = 0;
	p->left_image_height  = 0;

	p->max_ascent = p->max_descent = 0;
	p->prev_ascent = p->prev_descent = 0;

	return p;
}

static void
frameinfo_delete(FrameInfo *p)
{
	if (p) {
		aligninfolist_delete(p->align);
		htmlobjlist_delete(p->left);
		htmlobjlist_delete(p->right);
		free(p);
	}
}

StyleInfo *
styleinfo_new(Frame *f, TAGS tag)
{
	StyleInfo *p;
    if ((p = malloc(sizeof(*p))) == NULL) {
        perror("styleinfo_new");
        exit(1);
    }
	if (f->style) {
		p->size     = f->style->size;
		p->fonttype = f->style->fonttype;
		p->oc       = f->style->oc;
		p->pixel    = f->style->pixel;
	} else {
		p->size     = f->basefont - 1;
		p->fonttype = FONTTYPE_NORMAL;
		p->oc       = f->font[p->fonttype][p->size];
		p->pixel    = f->foreground;
	}
	p->tag  = tag;
	p->next = NULL;
	return p;
}

void
styleinfo_setcolor(Frame *f, StyleInfo *s, char *name)
{
	if (name)
		frame_alloc_color(f, name, &s->pixel);
}

void
styleinfo_setsize(Frame *f, StyleInfo *s, int size)
{
	s->size = size;
	s->oc = f->font[s->fonttype][s->size];
}

void
styleinfo_setfonttype(Frame *f, StyleInfo *s, int type)
{
	if (s->fonttype < FONTTYPE_PRE && type < FONTTYPE_PRE)
		s->fonttype |= type;
	else
		s->fonttype = type;
#if 0
	fprintf(stderr, "fonttype: %d\n", s->fonttype);
#endif
	s->oc = f->font[s->fonttype][s->size];
}

void
styleinfo_delete(Frame *f, StyleInfo *p)
{
	if (p)
		free(p);
}

Frame *
frame_new(Display *dpy, Window win,
		  FontInfoTableList fitl, 
		  FontNameTableList fntl,
		  FontNameTableList fntl_m,
		  u_int width, u_int height, int off_x, int off_y)
{
	int i;
	double size;

	Frame *p;
    if ((p = malloc(sizeof(*p))) == NULL) {
        perror("frame_new");
        exit(1);
    }
	
	p->dpy    = dpy;
	p->win    = win;
	p->width  = width;
	p->height = height;
	p->off_x  = off_x;
	p->off_y  = off_y;
	p->scr_x  = 0;
	p->scr_y  = 0;

	p->path = NULL;
	p->name = NULL;

	{
		XWindowAttributes attr;
		XGetWindowAttributes(p->dpy, p->win, &attr);
		p->colormap = attr.colormap;
		p->r_w = attr.width;
		p->r_h = attr.height;
		p->foreground = BlackPixelOfScreen(attr.screen);
		p->background = WhitePixelOfScreen(attr.screen);
		p->backup = XCreatePixmap(p->dpy, p->win, 
								  p->r_w, p->r_h, attr.depth);
		p->gc     = XCreateGC(p->dpy, p->win, 0, NULL);
		XSetGraphicsExposures(p->dpy, p->gc, False);
	}

	p->tagwork  = NULL;
	p->eof      = 0;

	p->bg_image = NULL;
	p->obj = p->line = p->line_p = NULL;

	p->links    = NULL;
	p->names    = NULL;

	p->images   = NULL;

	p->finfo = NULL;
	p->style = NULL;

	p->pixels = NULL;
	p->basefont = 3;

	size = 10;
	for (i=0;i<7;i++) {
#if 0
		fprintf(stderr, "make font: %.0f\n", size);
#endif
		p->font[FONTTYPE_NORMAL][i] = 
			CreateOC3(dpy, fitl, fntl, FONTTYPE_NORMAL, size);
		p->font[FONTTYPE_BOLD][i] = 
			CreateOC3(dpy, fitl, fntl, FONTTYPE_BOLD, size);
		p->font[FONTTYPE_ITALIC][i] =
			CreateOC3(dpy, fitl, fntl, FONTTYPE_ITALIC, size);
		p->font[FONTTYPE_BOLDITALIC][i] =
			CreateOC3(dpy, fitl, fntl, FONTTYPE_BOLDITALIC, size);
		p->font[FONTTYPE_PRE][i] = 
			CreateOC3(dpy, fitl, fntl_m, FONTTYPE_NORMAL, size);
		size *= 1.25;
	}

	return p;
}

static void
frame_clean(Frame *f)
{
	timerlist_delete(f->images);
	while (f->finfo)
		frame_pop_finfo(f);
	while (f->style)
		frame_pop_style(f);
	htmlobjlist_delete(f->obj);
	htmlhrefinfolist_delete(f->links);
	colorlistlist_delete(f, f->pixels);
	image_delete(f->bg_image);
}

void
frame_delete(Frame *f)
{
	int i, j;
	if (f) {
		frame_clean(f);
		XFreePixmap(f->dpy, f->backup);
		XFreeGC(f->dpy, f->gc);
		if (f->path)
			free(f->path);
		if (f->name)
			free(f->name);
		tagwork_delete(f->tagwork);
		free(f);
		for (i=0;i<7;i++)
			for (j=0;j<FONTTYPE_ALL;j++)
				DestroyOC(f->font[j][i]);
	}
}

void
frame_push_aligninfo(Frame *f, TAGS tag, AlignMode align)
{
	AlignInfo *p = aligninfo_new(tag, align);
	p->next = f->finfo->align;
	f->finfo->align = p;
}

void
frame_pop_aligninfo(Frame *f)
{
	AlignInfo *p = f->finfo->align;
	if (p) {
		f->finfo->align = p->next;
		aligninfo_delete(p);
	}
}

int
frame_find_aligninfo(Frame *f, TAGS tag)
{
	AlignInfo *p = f->finfo->align;
	while (p != NULL) {
		if (p->tag == tag)
			break;
		p = p->next;
	}
	return (p == NULL)?0:1;
}

void
frame_push_finfo(Frame *f, TAGS tag, int off_x, int off_y)
{
	FrameInfo *fi = frameinfo_new(f, off_x, off_y);
	fi->tag  = tag;
	fi->next = f->finfo;
	f->finfo = fi;
}

void
frame_pop_finfo(Frame *f)
{
	FrameInfo *oldfi;

	/* Ťե졼˲Ԥ */
	if (f->finfo) {
		oldfi = f->finfo;
		if (!oldfi->after_lf ||
			oldfi->left_image_margin ||
			oldfi->right_image_margin)
			frame_add_lf(f,BR_ALL);
		f->finfo = f->finfo->next;
		if (f->finfo) {
			frame_add_vspace(f, oldfi->cur_y);
			f->finfo->after_lf = 1;
			f->finfo->max_ascent  = 0;
			f->finfo->max_descent = 0;
			f->finfo->cur_x = 0;
		}
		frameinfo_delete(oldfi);
	}
}

int
frame_find_finfo(Frame *f, TAGS tag)
{
	FrameInfo *p = f->finfo;
	while (p != NULL) {
		if (p->tag == tag)
			break;
		p = p->next;
	}
	return (p == NULL)?0:1;
}

void
frame_push_style(Frame *f, StyleInfo *style)
{
	style->next = f->style;
	f->style = style;
}

void
frame_pop_style(Frame *f)
{
	if (f->style) {
		StyleInfo *olds = f->style;
		f->style = f->style->next;
		styleinfo_delete(f,olds);
	}
}

int
frame_find_style(Frame *f, TAGS tag)
{
	StyleInfo *p = f->style;
	while (p != NULL) {
		if (p->tag == tag)
			break;
		p = p->next;
	}
	return (p == NULL)?0:1;
}

void
frame_add_link(Frame *f, HTMLHrefInfo *p)
{
	p->next = f->links;
	f->links = p;
}

void
frame_add_obj(Frame *f, HTMLObj *obj)
{
	FrameInfo *fi = f->finfo;
	obj->x = fi->off_x + CUR_X;
	obj->y = fi->off_y + fi->cur_y;
	fi->max_ascent  = MAX(fi->max_ascent, obj->ascent);
	fi->max_descent = MAX(fi->max_descent, obj->descent);
	fi->cur_x += obj->width;
	f->formatted_width = MAX(f->formatted_width, 
							 obj->x + obj->rbearing);
	obj->next = f->obj;
	obj->prev = f->obj->prev;
	f->obj->prev->next = obj;
	f->obj->prev = obj;
}

void
frame_add_obj_before_line(Frame *f, HTMLObj *obj)
{
	if (f->line == f->obj) {	/* special case */
		obj->next = f->obj;
		obj->prev = f->obj;
		f->obj->next = obj;
		f->obj->prev = obj;
		f->line = obj;
	} else {
		obj->next = f->line;
		obj->prev = f->line->prev;
		f->line->prev->next = obj;
		f->line->prev = obj;
	}
}

void
frame_add_left(Frame *f, HTMLObj *obj)
{
	FrameInfo *fi = f->finfo;
	obj->next = fi->left;
	obj->prev = fi->left->prev;
	fi->left->prev->next = obj;
	fi->left->prev = obj;
}

void
frame_add_right(Frame *f, HTMLObj *obj)
{
	FrameInfo *fi = f->finfo;
	obj->next = fi->right;
	obj->prev = fi->right->prev;
	fi->right->prev->next = obj;
	fi->right->prev = obj;
}

/* ζɲ */
void
frame_add_vspace(Frame *f, int linesep)
{
	FrameInfo *fi = f->finfo;

	fi->cur_y += linesep;
	
	if (fi->left_image_height > 0) {
		fi->left_image_height -= linesep;
		if (fi->left_image_height <= 0) 
			fi->left_image_margin = 0;
	}

	/*  */
	while (fi->left_image_height < 0 
		   || (fi->left_image_height == 0 && fi->left->next != fi->left)) {
		int line = fi->left_image_height;
		if (fi->left->next != fi->left) {
			HTMLObj *p;

			p = fi->left->next;
			fi->left_image_margin = p->width;
			fi->left_image_height = p->height;
			p->next->prev = fi->left;
			fi->left->next = p->next;

			p->x = 0;
			p->y = fi->off_y + fi->cur_y + line;
			frame_add_obj_before_line(f,p);

			fi->left_image_height += line;
		} else {
			fi->left_image_margin = 0;
			fi->left_image_height = 0;
			break;
		}
	}

	/*  */
	if (fi->right_image_height > 0) {
		fi->right_image_height -= linesep;
		if (fi->right_image_height <= 0) 
			fi->right_image_margin = 0;
	}
	
	while (fi->right_image_height < 0 
		   || (fi->right_image_height == 0 && fi->right->next != fi->right)) {
		int line = fi->right_image_height;
		if (fi->right->next != fi->right) {
			HTMLObj *p;

			p = fi->right->next;
			fi->right_image_margin = p->width;
			fi->right_image_height = p->height;
			p->next->prev = fi->right;
			fi->right->next = p->next;

			p->x = fi->off_x + fi->width - fi->right_image_margin;
			p->y = fi->off_y + fi->cur_y + line;
			frame_add_obj_before_line(f,p);

			fi->right_image_height += line;
		} else {
			fi->right_image_margin = 0;
			fi->right_image_height = 0;
			break;
		}
	}
}

/*  & ƼƷ׻ */
void
frame_add_lf(Frame *f, int mode)
{
	HTMLObj *o;
	FrameInfo *fi = f->finfo;
	int linesep;
	int dx = 0;

	if (fi->cur_x < WIDTH) {
		switch (fi->align->align) {
		case ALIGN_RIGHT:
			dx = WIDTH - fi->cur_x;
			break;
		case ALIGN_CENTER:
			dx = (WIDTH - fi->cur_x) / 2;
			break;
		default:
		case ALIGN_LEFT:
			dx = 0;
			break;
		}
	}

	/* ιԤΤ٤Ƥʸ򤺤餹 */
	o = f->line->next;
	while (o != f->obj) {
		o->y += fi->max_ascent;
		o->x += dx;
		o = o->next;
	}
	f->line = f->obj->prev;
	f->line_p = f->line->prev;

	linesep = fi->max_descent + fi->max_ascent + 10; /* XXX */
	fi->cur_x = 0;
	fi->max_ascent = 0;
	fi->max_descent = 0;

 retry:
	frame_add_vspace(f, linesep);
	
	/* ҤԤ */
	if (mode == BR_ALL) {
		if ((fi->right_image_height > 0 || fi->left_image_height > 0) ) {
			linesep = MAX(fi->right_image_height, fi->left_image_height) + 10;
			goto retry;
		}
	} else if (mode == BR_RIGHT) {
		if (fi->right_image_height > 0) {
			linesep = fi->right_image_height + 10;
			goto retry;
		}
	} else if (mode == BR_LEFT) {
		if (fi->left_image_height > 0) {
			linesep = fi->left_image_height + 10;
			goto retry;
		}
	}

#if 0
	fprintf(stderr, "margin info: %d %d %d %d\n", 
			fi->width, fi->left_image_margin, fi->right_image_margin,
			f->formatted_width);
#endif
	fi->after_lf = 1;
}

void
frame_add_body(Frame *f)
{
	if (f->bg_image && f->bg_image->pixmap != None) {
		XSetWindowBackgroundPixmap(f->dpy, f->win, f->bg_image->pixmap);
		XSetTile(f->dpy, f->gc, f->bg_image->pixmap);
		XSetTSOrigin(f->dpy, f->gc, f->off_x - 
					 f->scr_x, f->off_y -f->scr_y);
		XSetFillStyle(f->dpy, f->gc, FillTiled);
		XFillRectangle(f->dpy, f->backup, f->gc, 0, 0, f->r_w, f->r_h);
		XSetFillStyle(f->dpy, f->gc, FillSolid);
	} else {
		XSetWindowBackground(f->dpy, f->win, f->background);
		XSetForeground(f->dpy, f->gc, f->background);
		XFillRectangle(f->dpy, f->backup, f->gc, 0, 0, f->r_w, f->r_h);
	}
	frame_expose(f, 0, 0, f->r_w, f->r_h);
}

#define BULLET_SIZE 8

void
frame_add_bullet(Frame *f, int type)
{
	HTMLObj *obj;
	obj = htmlobj_new();
	obj->type = HTMLOBJ_BULLET;
	obj->data = (void*)type;

	obj->lbearing = -BULLET_SIZE/2;
	obj->rbearing = BULLET_SIZE/2;
	obj->width    = BULLET_SIZE;
	obj->height   = BULLET_SIZE;
	obj->ascent   = BULLET_SIZE;
	obj->descent  = 0;
	
	frame_add_obj(f,obj);
}


/* ʸɲ */
void
frame_add_string(Frame *f, wchar_t *str, size_t len, int mode)
{
	XCharStruct all;
	HTMLObj *obj;
	HTMLTextObj *t;

	FrameInfo *fi = f->finfo;

	if (len == 0) 
		return;

	obj = htmlobj_new();
	t   = htmltextobj_new();
	obj->type = HTMLOBJ_TEXT;
	obj->data = t;

	t->oc      = f->style->oc;

	if (f->linkmode) {
		obj->link  = f->links;
		obj->pixel = f->link_style->pixel;
	} else {
		obj->pixel = f->style->pixel;
	}
		
	t->str = str;
	t->len = len;

	WcExtents(f->style->oc, str, len, &all);
	obj->lbearing = all.lbearing;
	obj->rbearing = all.rbearing;
	obj->width    = all.width;
	obj->height   = all.ascent + all.descent;
	obj->ascent   = all.ascent;
	obj->descent  = all.descent;

	fi->prev_ascent  = all.ascent;
	fi->prev_descent = all.descent;
	
	if (mode)
		fi->cur_x -= obj->width;
	frame_add_obj(f,obj);
	
#if 0
	{
		int i;
		fprintf(stderr,"add string:[%d,%d]", obj->x, obj->y);
		for (i=0;i<len;i++)
			fprintf(stderr, "[%x]", str[i]);
		fprintf(stderr,"\n");
	}
#endif

}

void
frame_add_image(Frame *f,
				Animate *a, int w, int h, int b, int align)
{
	int ascent = 0;
	int height = 0;
	FrameInfo *fi = f->finfo;
	HTMLObj *obj;
	HTMLImageObj *i;

	obj = htmlobj_new();
	i = htmlimageobj_new();
	i->anim = a;
	obj->type = HTMLOBJ_IMAGE;
	obj->data = i;

	if (f->linkmode) {
		obj->link  = f->links;
		obj->pixel = f->link_style->pixel;
	} else
		obj->pixel = f->style->pixel;

	if (i->anim != NULL) {
		obj->width    = i->anim->w;
		obj->rbearing = i->anim->w;
		obj->height   = height = i->anim->h;
		/* entry animation pattern */
		if (i->anim->list->next->next != i->anim->list)
			timer_entry(f->images, obj);
	}
	
	switch (align) {
	case ALIGN_LEFT:
	case ALIGN_RIGHT:
		ascent = 0;
		break;
	case ALIGN_TOP:
		ascent  = fi->prev_ascent;
		break;
	case ALIGN_BOTTOM:
		ascent = height;
		break;
	case ALIGN_ABSMIDDLE:
		ascent = (height + fi->prev_ascent - fi->prev_descent)/2;
		break;
	case ALIGN_NONE:
	case ALIGN_MIDDLE:
	default:
		ascent = height / 2;
		break;
	}
	
	obj->ascent  = ascent;
	obj->descent = obj->height - ascent;

	if (align == ALIGN_LEFT) {
		if (fi->after_lf && fi->left_image_height == 0) {
			fi->left_image_margin = obj->width;
			fi->left_image_height = obj->height;
			obj->x = 0;
			obj->y = fi->off_y + fi->cur_y;
			frame_add_obj_before_line(f,obj);
		} else							/* ꥹȤ˥ȥ */
			frame_add_left(f,obj);
	} else if (align == ALIGN_RIGHT) {
		if (fi->cur_x < (fi->width - fi->left_image_margin - obj->width)
			&& fi->right_image_margin == 0) {
			fi->right_image_margin = obj->width;
			fi->right_image_height = obj->height;
			obj->x = fi->off_x + fi->width - fi->right_image_margin;
			obj->y = fi->off_y + fi->cur_y;
			frame_add_obj_before_line(f,obj);
		} else							/* ꥹȤ˥ȥ */
			frame_add_right(f,obj);
	} else {
		if (f->finfo->pre != TAG_OPEN && fi->cur_x + obj->width >= WIDTH)
			frame_add_lf(f,BR_NORMAL);
		frame_add_obj(f,obj);
		f->finfo->after_lf = 0;
	}
}

void
frame_add_hr(Frame *f, int align, int width, int size)
{
	HTMLObj *obj = htmlobj_new();

	obj->type = HTMLOBJ_HR;
	obj->ascent = size/2;
	obj->descent = size - obj->ascent;
	obj->width = obj->rbearing = width;
	obj->height = size;
	obj->pixel  = f->style->pixel;

	frame_add_obj(f,obj);

	frame_push_aligninfo(f,0,align);
	frame_add_lf(f, BR_NORMAL);
	frame_pop_aligninfo(f);
}

void
frame_expose(Frame *f, int x, int y, int w, int h)
{
	XCopyArea(f->dpy, f->backup, f->win, f->gc,
			  x, y, w, h, x, y);
}

void
frame_expose_region(Frame *f, Region reg)
{
	XSetRegion(f->dpy, f->gc, reg);
	XCopyArea(f->dpy, f->backup, f->win, f->gc, 0, 0, f->r_w, f->r_h, 0, 0);
	XSetClipMask(f->dpy, f->gc, None);
}

static void
frame_init_params(Frame *f)
{
	f->linkmode = 0;
	f->formatted_width  = 0;
	f->formatted_height = 0;
	f->scr_x  = 0;
	f->scr_y  = 0;

	f->link_style = styleinfo_new(f,0);
	styleinfo_setcolor(f,f->link_style,"red");
	f->links     = NULL;
	f->names     = NULL;

	frame_push_finfo(f, 0, 0, 0);
	frame_push_style(f, styleinfo_new(f,0));
	f->obj = f->line = f->line_p = htmlobj_new();

	f->images = timer_new();

	f->bg_image = NULL;
	f->pixels   = NULL;
	f->basefont = 3;

	f->eof = 0;
}

void
frame_resize(Frame *f, int width, int height, int off_x, int off_y)
{
	frame_clean(f);

	f->width  = width;
	f->height = height;
	f->off_x  = off_x;
	f->off_y  = off_y;
	f->scr_x  = 0;

	frame_init_params(f);
	tagwork_reset(f->tagwork);

	{
		XWindowAttributes attr;
		XGetWindowAttributes(f->dpy, f->win, &attr);
		if (f->r_w != attr.width 
			|| f->r_h != attr.height) {
			f->r_w = attr.width;
			f->r_h = attr.height;
			XFreePixmap(f->dpy, f->backup);
			f->backup = XCreatePixmap(f->dpy, f->win, 
									  f->r_w, f->r_h, attr.depth);
		}
	}
}

int
frame_open(Frame *f, int fd, char *name)
{
	char *p;
	int len;

 	frame_clean(f);
	frame_init_params(f);

	if (f->path)
		free(f->path);
	if (f->name)
		free(f->name);

	if (name == NULL)
		name = "./";

	if ((p = strrchr(name,'/'))) {
		p++;
		f->name = strdup(p);
		if ((f->path = malloc((len=(p - name))+1)) == NULL) {
			perror("tagwork_open");
			exit(1);
		}
		strncpy(f->path, name, len);
		f->path[len] = '\0';
	} else {
		f->path = NULL;
		f->name = strdup(name);
	}

	f->fd = fd;
	tagwork_delete(f->tagwork);
	f->tagwork = tagwork_open(f->fd);

	frame_add_body(f);

	return f->fd;
}

FILE *
frame_file_open(Frame *f, char *name)
{
	FILE *fp;
	char *buf;
	if (*name == '/' || f->path == NULL) {
		fp = fopen(name, "r");
	}else {
		if ((buf = malloc(strlen(name) + strlen(f->path) + 1)) == NULL)
			return NULL;
		sprintf(buf, "%s%s", f->path, name);
		fp = fopen(buf, "r");
		free(buf);
	}
	return fp;
}
