#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include "xwchar.h"

#include "tag.h"
#include "msgbuf.h"
#include "animate.h"

/* TagOpt */

static TagOpt*
tagopt_new(void)
{
	TagOpt *p;
	if ((p = malloc(sizeof(*p))) == NULL) {
		perror("tagopt_new");
		exit(1);
	}
	p->next = NULL;
	p->oi   = NULL;
	p->val  = NULL;
	return p;
}

static void
tagopt_delete(TagOpt *p)
{
	if (p) {
		if (p->val)
			free(p->val);
		free(p);
	}
}

static void
tagoptlist_delete(TagOptList l)
{
	TagOpt *p = l, *q;
	while (p != NULL) {
		q = p;
		p = p->next;
		tagopt_delete(q);
	}
}

#ifdef DEBUG
static void
tagopt_showinfo(TagOpt *p)
{
	if (p->oi) {
		fprintf(stderr, "option:%s", p->oi->name);
		if (p->val)
			fprintf(stderr, "[%s]\n", p->val);
		else
			fprintf(stderr, "\n");
	}
}
#endif

/* Tag */

static Tag*
tag_new(void)
{
	Tag *p;
	if ((p = malloc(sizeof(*p))) == NULL) {
		perror("tag_new");
		exit(1);
	}
	p->next = p->prev = p;
	p->type = TAG_none;
	p->ti   = NULL;
	p->data = NULL;
	p->opt  = NULL;
	p->mode = 0;
	return p;
}

static void
tag_delete(Tag *p)
{
	if (p) {
		tagoptlist_delete(p->opt);
		if (p->data) {
			if (p->type == TAG_normal && p->ti->id == TAG_img)
				animate_delete(p->data);
			else
				free(p->data);
		}
	}
}


static void
taglist_delete(TagList l)
{
	Tag *p, *q;
	if (l) {
		p = l->next;
		while (p != l) {
			q = p;
			p = p->next;
			tag_delete(q);
		}
		tag_delete(l);
	}
}

#ifdef DEBUG
void
tag_showinfo(Tag *p)
{
	fprintf(stderr,"[%p]: ",p);
	if (p->type == TAG_normal) {
		fprintf(stderr, "TAG:%s [%s]\n", p->ti->name, p->mode?"close":"open");
		if (p->opt) {
			TagOpt *o = p->opt;
			while (o != NULL) {
				tagopt_showinfo(o);
				o = o->next;
			}
		}
	} else if (p->type == TAG_string) {
		wchar_t *w = p->data;
		fprintf(stderr, "TEXT DATA\n");
		if (w != NULL)
			while (*w)
				fprintf(stderr, "[%lx]", *w++);
		fprintf(stderr, "\n");
	} else
		fprintf(stderr, "\n");
}
#endif

#define BUFSIZE 2048

static wchar_t
tagwork_getchar(TagWork *w)
{
	static wchar_t prev = 0;
	wchar_t ret;
	int len, l;
 retry:
	if (w->eof >= 2)
		return 0;
	
	if (!w->eof && ((len = (w->p2 - w->p)) < (BUFSIZE/2))) {
		memcpy(w->data, w->p, len);
		w->p  = w->data;
		w->p2 = w->p + len;
		if ((l = read(w->fd, w->p2, BUFSIZE - len)) > 0) {
			w->p2 += l;
		} else if (l < 0) {
			perror("read");
			exit(1);
		} else if (l == 0) {
			*(w->p2) = '\0';
			w->eof = 1;
		}
	}

	len = mbrtowc(&ret, w->p, _mb_cur_max_, &w->state);
	if (ret)
		w->p += len;
	else 
		w->eof = 2;

#if 0
	fprintf(stderr, "[%lx]", ret);
#endif
	if (w->pre == TAG_CLOSE) {	/* format mode */
		if (ret == '\r' || ret == '\n') {
			if (IS_MB(prev))
				goto retry;
			else
				ret = ' ';		/* ¾ξˤ */
		} 
		if (w->tagend && IS_MB(ret) && ((ret & 0x7f7f) == 0x2121))
			goto retry;			/* ƬѶ */
		if ((w->tagend && ret == ' ')
			|| (ret == ' ' && prev == ' '))	/* ƬorϢ³Ϻ */
			goto retry;
	} else { /* preformatted mode */
		/* mac:CR, dos:CRLF, unix:LF  --> LF */
		if (ret == '\n' && prev == '\r') {
			prev = '\n';
			goto retry;
		}
		if (ret == '\r') {
			prev = ret;
			return '\n';
		}
	}
	w->tagend = 0;
	prev = ret;

		
	return ret;
}

static wchar_t
space_skip(TagWork *w, int ch)
{
	if ((!ch) || ch > ' ') return ch;
	while ((ch = w->GetChar(w))) if (ch > ' ') break;
	return ch;
}

static Tag*
tagwork_tagget(TagWork *w)
{
	char work[100], *tmp;
	int len;
	wchar_t ch;
	Tag *tag;
	TagOpt *last, *add;

	tag = tag_new();
	
	if ((ch = w->GetChar(w)) == '/') {
		tag->mode = 1;	/* close tag */
		ch = w->GetChar(w);
	}
			
	tmp = work;
	len = 0;
	do {
		if (ch == '>' || ch <= ' ' || ch > 0xff ) break;
		*tmp++ = tolower(ch);
		if (++len >= sizeof(work)-1)
			goto abort;
	} while ( (ch = w->GetChar(w)) );
	*tmp = '\0';
	
	/* comment */
	if (!strncmp(work,"!--",3)) {
		while ( (ch = w->GetChar(w)) )
			if (ch == '>') break;
		if (!ch) goto abort;
		tag->type = TAG_comment;
		return tag;
	}

	if (*work == '!') {
		while ( (ch = w->GetChar(w)) )
			if (ch == '>') break;
		if (!ch) goto abort;
		tag->type = TAG_unknown;
		return tag;
	}

	tag->type = TAG_normal;
	if ((tag->ti = GetTagInfo(work)) == NULL)
		tag->type = TAG_unknown;

	if (ch == '>') return tag;

	/* options */
    do {
        if (!(ch = space_skip(w,ch))) goto abort;
        tmp = work;
		len = 0;
        do {
            if (ch =='>' || ch == '=' || ch <= ' ' || ch > 0xff) break;
            *tmp++ = tolower(ch);
			if (++len >= sizeof(work)-1)
				goto abort;
        } while ((ch = w->GetChar(w)));
        if (!ch) goto abort;



        *tmp = '\0';
        if (!strlen(work)) {
            if (ch == '>') break;
            if (ch == '=') ch = w->GetChar(w);
            continue;
        }
		
		add = tagopt_new();

		add->oi = GetOptInfo(work);

		if (!tag->opt) {
			tag->opt = add;
		} else {
			last->next = add;
		}
		last = add;

        if (!(ch = space_skip(w,ch))) goto abort;

        /* ͤĤץ󤫡 */
        if (ch == '=') {
            tmp = work;
			len = 0;
            if (!(ch = space_skip(w,' '))) goto abort;

            if (ch == '"'){
                while ((ch = w->GetChar(w))) {
                    if (ch == '"') break;
                    *tmp++ = ch;
					if (++len >= sizeof(work)-1)
						goto abort;
                }
                if (ch == '"') ch = w->GetChar(w);
            } else
                do {
                    if (ch == '>' || ch <= ' ') break;
                    *tmp++ = ch;
					if (++len >= sizeof(work)-1)
						goto abort;
                } while ((ch = w->GetChar(w)));

            if (!ch) goto abort;
            *tmp= '\0';
            if (add->oi)
                add->val = strdup(work);
        }
    } while( ch != '>' );
    return tag;

 abort:
	tag_delete(tag);
	return NULL;
}

static void
tagwork_addtag(TagWork *w, Tag *tag)
{	
	tag->next = w->tags;
	tag->prev = w->tags->prev;
	w->tags->prev->next = tag;
	w->tags->prev = tag;
}

/* λ */
#define ENTITYLEN 64
static int
entity(TagWork *w)
{
    int c, i, val;
    char buf[ENTITYLEN];
    EntList *ep;
    wchar_t wc;

    for (i = 0; i < ENTITYLEN ; i++) {
		wc = w->GetChar(w);
		if (!IS_ASCII(wc))
			break;
		c = wc & 0x7f;

        if (c == ';' || isspace(c)) {
			buf[i] = '\0';
			if (buf[0] == '#' && (val = atoi(buf + 1)) > 0) {
				return val;
			} else if ((ep = GetEntity(buf))) {
				return ep->code;
			}
			break;
		}
		buf[i] = c;
    }
    return 0;   /* signifies unknown entity name */
}

static int
tagwork_parse(TagWork *w)
{
	int ret = 0;
	wchar_t ch;
	wchar_t *mp;
	Tag *tag = NULL;

	MessageBuffer *mes = msgbuf_new();
	
	for(;;) {

		if (!(ch = w->GetChar(w))) {	/* eof */
#if 0
			fprintf(stderr, "tagwork eof!!\n");
#endif
			ret = 1;
			break;
		}

		/* tag */
		if (ch == '<') {
			if ((mp = msgbuf_cut(mes))) {
				tag = tag_new();
				tag->type = TAG_string;
				tag->data = mp;
				tagwork_addtag(w,tag);
			}

			if (!(tag = tagwork_tagget(w))) break;

			if (tag->type == TAG_normal && tag->ti->id == TAG_pre)
				w->pre = tag->mode;

			tagwork_addtag(w,tag);
			w->tagend = 1;
			break;
		}

		/* text start */

		/* λ */
		if (ch == '&')
			ch = entity(w);
		
		msgbuf_add(mes, ch);
	}

	if ((mp = msgbuf_cut(mes))) {
		tag = tag_new();
		tag->type = TAG_string;
		tag->data = mp;
		tagwork_addtag(w,tag);
	}

	msgbuf_delete(mes);

	return ret;
}

void InitTagInfo(void);
void InitOptInfo(void);
void InitEntities(void);

static void
tagwork_init(void)
{
	static int flag = 0;
	if (!flag) {
		InitOptInfo();
		InitTagInfo();
		InitEntities();
		flag = 1;
	}
}

static TagWork *
tagwork_new(void)
{
	TagWork *p;
	if ((p = malloc(sizeof(*p))) == NULL) {
		perror("tagwork_new");
		return NULL;
	}

	if ((p->data = malloc(BUFSIZE+1)) == NULL) {
		perror("togwork_new");
		exit(1);
	}
	
	p->fd  = -1;
	p->eof = 0;

	p->p = p->p2 = p->data;
	mbrtowc(NULL,NULL,0,&p->state);	/* init state */

	p->tags    = p->cur = tag_new();
	p->main    = tagwork_parse;
	p->GetChar = tagwork_getchar;
	
	p->pre = TAG_CLOSE;
	p->tagend = 1;
	
	tagwork_init();
	return p;
}

Tag *
tagwork_gettag(TagWork *w)
{
	if (w->cur->next == w->tags)
		return NULL;
	w->cur = w->cur->next;
	return w->cur;
}

TagWork *
tagwork_open(int fd)
{
	TagWork *w = tagwork_new();
	w->fd = fd;
	return w;
}

void
tagwork_delete(TagWork *tw)
{
	if (tw) {
		if (tw->data)
			free(tw->data);
		taglist_delete(tw->tags);
		free(tw);
	}
}

void
tagwork_reset(TagWork *tw)
{
	if (tw) 
		tw->cur = tw->tags;
}
